#include <stdio.h>
#include <string.h>

#include "rc2.h"

//void print_key(unsigned char *key)
//{
//    short   i;
//    for (i = 0; i < 128; i+= 2) {
//	printf("%02x%02x ", key[i], key[i+1]);
//	if (i % 32 == 30)
//	    printf("\n");
//    }
//    printf("\n");
//}

void usage(char *progname)
{
    fprintf(stderr,
	"usage: %s <ciphertext> <key> <trials>\n"
	"  ciphertext - name of file containing IV and ciphertext.  first 8\n"
	"      bytes of file should be IV and the remaining data should be\n"
	"      data encrypted with 40-bit RC2 in CBC-mode.\n"
	"  key - 10-digit hexidecimal starting value.\n"
	"  trials - number of keys to try, must be even.\n",
	progname);
}

unsigned char hex(char c)
{
    if (c >= '0' && c <= '9')
	return c - '0';
    if (c >= 'A' && c <= 'Z')
	return c - 'A' + 10;
    if (c >= 'a' && c <= 'z')
	return c - 'a' + 10;
    return 0xFF;
}

#define xorbuf(d, s) \
do { \
    (d)[0] ^= (s)[0]; \
    (d)[1] ^= (s)[1]; \
    (d)[2] ^= (s)[2]; \
    (d)[3] ^= (s)[3]; \
} while (0)

#define is_text(p) \
(((unsigned char*)(p))[0] >= 128 ? 0 : \
 ((unsigned char*)(p))[1] >= 128 ? 0 : \
 ((unsigned char*)(p))[2] >= 128 ? 0 : \
 ((unsigned char*)(p))[3] >= 128 ? 0 : \
 ((unsigned char*)(p))[4] >= 128 ? 0 : \
 ((unsigned char*)(p))[5] >= 128 ? 0 : \
 ((unsigned char*)(p))[6] >= 128 ? 0 : \
 ((unsigned char*)(p))[7] >= 128 ? 0 : 1)

int main(int argc, char **argv)
{
    /*
     * odd_offset_key has 259 elements so we have a dummy slot at the end
     * and two dummy slots at the front.
     */
    unsigned char   odd_offset_key[257], even_offset_key[257];

    unsigned char   *plaintext, *ciphertext, IV[8], *end;
    unsigned char   *oo, *eo, *oe, *ee, key_0[5], buffer[8];
    unsigned short  *p, *c, *tp, *tc;
    long    l, trials;
    short   i, j, match;
    FILE    *input;
    long    len;

    if (argc != 4) {
	usage(argv[0]);
	return -1;
    }

    /* open in the input file */
    if ((input = fopen(argv[1], "rb")) == NULL) {
	fprintf(stderr, "error opening %s for input.\n", argv[1]);
	perror(argv[0]);
	return -1;
    }

    /* get the file length minus the IV */
    fseek(input, 0L, SEEK_END);
    len = ftell(input) - 8;
    rewind(input);

    /* read the IV and ciphertext */
    if (fread(IV, 1, 8, input) != 8) {
	fprintf(stderr, "error reading from %s.\n", argv[1]);
	perror(argv[0]);
	return -1;
    }
    if ((ciphertext = (unsigned char*)malloc(len)) == NULL ||
	(plaintext  = (unsigned char*)malloc(len)) == NULL) {
	fprintf(stderr, "error allocating memory.\n");
	return -1;
    }
    if (fread(ciphertext, 1, len, input) != len) {
	fprintf(stderr, "error reading from %s.\n", argv[1]);
	perror(argv[0]);
	return -1;
    }
    c = (unsigned short*)ciphertext;
    p = (unsigned short*)plaintext;

    /* load the starting key and the number of keys to try */
    if (strlen(argv[2]) != 10) {
	fprintf(stderr, "invalid starting key.\n");
	usage(argv[0]);
	return -1;
    }
    for (i = 0; i < 5; i++)
	key_0[i] = (hex(argv[2][2*i]) << 4) | hex(argv[2][2*i+1]);

    trials = atol(argv[3]);
    if (trials % 2 == 1 || trials < 0) {
	fprintf(stderr, "invalid number of trials.\n");
	usage(argv[0]);
	return -1;
    }

    /* need to have two different key schedules since two successive key 
     * key schedules have different alignments.  the two data structures
     * will make it so that the current key schedule always has proper
     * word alignment so we can cast to a (unsigned short *).
     */
    memset(odd_offset_key,  0, sizeof(odd_offset_key));
    memset(even_offset_key, 0, sizeof(even_offset_key));

    /* oo and eo point to the current key schedule.  they effectively
     * represent a sliding window through odd_offset_key and
     * even_offset_key respectively.  once oo and eo pass through oe
     * and ee respectively, then they are reset to just past the
     * beginning of their respective buffers.  as a result we must
     * always make sure that we update the bytes in the lower half
     * of the buffers so that they will be correct when we reset oo
     * and eo.
     */
    oo = odd_offset_key + 2;
    eo = even_offset_key + 2;
    oe = oo + 128;
    ee = eo + 128;

    memcpy(eo, key_0, sizeof(key_0));
    rc2_full_keyschedule(eo);
    memcpy(oo, eo+1, 127);
    rc2_partial_keyschedule(oo);

    end = ciphertext + len;

    /* perform trials until we find a key or run out of keys */
    for (l = 0, trials /= 2; l < trials; l++) {
	/* perform a trial encryption with the key in even_offset_key */
	rc2_decrypt((unsigned short*)eo, c, p);
	xorbuf(p, IV);
	for (tp = p, tc = c, match = 0; match < 5; match++) {
	    if (is_text(tp) == 0)
		goto fail_1;
	    tp += 4;
	    tc += 4;
	    rc2_decrypt((unsigned short*)eo, tc, tp);
	    xorbuf(tp, tc-4);
	}

	/* almost definitely found a match! */
	*(tp+4) = 0;
	if (strstr((char*)p, "Content-") != NULL) {
	    fprintf(stderr, "Very high probability match.\n");
	    fprintf(stderr, "key = %02x%02x%02x%02x%02x.\n",
		inverse[eo[0]], eo[1], eo[2], eo[3], eo[4]);
	    fprintf(stderr, "first 40 bytes of plaintext.\n%s\n", plaintext);
	    return 1;
	} else {
	    fprintf(stderr, "High probability match.\n");
	    fprintf(stderr, "key = %02x%02x%02x%02x%02x.\n",
		inverse[eo[0]], eo[1], eo[2], eo[3], eo[4]);
	    fprintf(stderr, "first 40 bytes of plaintext.\n%s\n", plaintext);

	    /* keep on searching */
	}

	fail_1:

	/* the next key for even_offset_key comes two iterations later in
	 * the cycle than the current one, *not* one.  so get the info on
	 * the next iteration from odd_offset_key and compute the second
	 * iteration.
	 */
	eo[0] = eo[128] = oo[127];

	/* check if we need to wrap back to the beginning of the buffer */
	if ((eo += 2) == ee)
	    eo = even_offset_key + 2;
	rc2_partial_keyschedule(eo);

	/* update in the event we wrap back to the beginning */
	eo[-1] = eo[127];

	/* perform a trial decryption with the key in odd_offset_key */
	rc2_decrypt((unsigned short*)oo, c, p);
	xorbuf(p, IV);
	for (tp = p, tc = c, match = 0; match < 5; match++) {
	    if (is_text(tp) == 0)
		goto fail_2;
	    tp += 4;
	    tc += 4;
	    rc2_decrypt((unsigned short*)oo, tc, tp);
	    xorbuf(tp, tc-4);
	}

	/* almost definitely found a match! */
	*(tp+4) = 0;
	if (strstr((char*)p, "Content-") != NULL) {
	    fprintf(stderr, "Very high probability match.\n");
	    fprintf(stderr, "key = %02x%02x%02x%02x%02x.\n",
		inverse[oo[0]], oo[1], oo[2], oo[3], oo[4]);
	    fprintf(stderr, "first 40 bytes of plaintext.\n%s\n", plaintext);
	    return 1;
	} else {
	    fprintf(stderr, "High probability match.\n");
	    fprintf(stderr, "key = %02x%02x%02x%02x%02x.\n",
		inverse[oo[0]], oo[1], oo[2], oo[3], oo[4]);
	    fprintf(stderr, "first 40 bytes of plaintext.\n%s\n", plaintext);

	    /* keep on searching */
	}

	fail_2:

	/* the next key for odd_offset_key comes two iterations later in
	 * the cycle than the current one, *not* one.  so get the info on
	 * the next iteration from even_offset_key and compute the second
	 * iteration.
	 */
	oo[0] = oo[128] = eo[127];

	/* check if we need to wrap back to the beginning of the buffer */
	if ((oo += 2) == oe)
	    oo = odd_offset_key + 2;
	rc2_partial_keyschedule(oo);

	/* update in the event we wrap back to the beginning */
	oo[-1] = oo[127];
    }

    //printf("failed.\n");
   fprintf(stderr, "next key in cycle = %02x%02x%02x%02x%02x.\n",
		inverse[eo[0]], eo[1], eo[2], eo[3], eo[4]);

    return 0;
}

